function [ U, opts ] = pdecol( f_func, dbdu_func, dbdux_func, dzdt_func, uinit_func, varargin )
% PDECOL - Solve one dimensional PDE's using ALGORITHM 540 PDECOL.
% 
% The types of PDE's that can be solved take the following form
% 
% u_t = f(t,x,u,u_x,u_xx)
% 
% With u a vector of NPDE unknown functions, u_t and u_x representing time
% and spacial derivatives respectfully.  Boundary conditions take the form
% of equations of the following form
% 
% b(u,u_x) = z(t)
% 
% and must be consistent with the initial conditions.  The user of the 
% routine must at minimum specify the following octave functions:
% 
% A function to evaluate f (defined above)
% A function to evaluate the derivative of b with respect to u
% " """""""" "" """""""" """ """""""""" "" " """" """"""" "" u_x
% " """""""" "" """""""" """ """""""""" "" z """" """"""" "" t
% A function to evaluate and return the initial conditions
% 
% This method in space is a finite element collocation method with piecewise 
% polynomials (B-Splines).  The time integration is performed using the method
% of lines with choices for several standard ordinary differential equations 
% solvers.  Many more details can be found in the original article:
% 
% Algorithm 540: PDECOL, General Collocation Software for Partial Differential Equations
% by N. K. Madsen, R. F. Sincovec
% ACM Transactions on Mathematical Software (TOMS), Volume 5 Issue 3 (September 1979)
% 
% Written by:
% -- 
% John L. Weatherwax                2005-04-27
% 
% email: wax@alum.mit.edu
% 
% Please send comments and especially bug reports to the
% above email address.
% 
%-----

if( nargin < 5 ) error( 'Not enough input arguments' ); end

% Default pdecol input arguments: 
%  

% The initial time:
%
t0 = 0.0;

% The left end of the computational domain:
% 
xL = 0;        

% The right end of the computational domain:
% 
xR = 1;        

% The number of intervals to divide the domain into:
% 
nint = 30; 

% The initial stepsize in time:
% 
dt = 1e-7;     

% The desired relative accuracy:
% 
eps = 1e-4;     

% The order of the piecewise polynomial B-Spline space to be used (ranging from 3-20):
% 
kord = 4;        

% The number of required breakpoint continuity conditions (rangind from 1-kord):
% 
ncc = 2;        

% The array of piecewise polynomial breakpoints (not required to be uniformly spaced):
xbkpt = linspace( xL, xR, nint+1 ); 

% Output times the solution is desired:
% 
tGrid = [ t0 t0+1.0 ]; 

% Algorithm Method Flags:
% 
% MF = 10*METH + MITER

% METH  = 1 Adams method (a generalization of Crank-Nicolson)
%       = 2 backwards differentiation formulas (bdf) or stiff method of Gear
% 
METH = 2; 

% MITER = 1 chord method with analytic jacobians, user supplies subroutines for the derivative evaluation
%       = 2 chord method calculated via finite differences
%
MITER = 2; 

% mf={12,22} for approximate derivatives or {11,21} for exact derivatives
mf = 10*METH + MITER;

% Override default arguments with "varargin" inputs (if any):
%
i=1; 
while i < length(varargin)
  arg = varargin{i};
  switch lower(arg)
   case 't0'
    t0 = varargin{i+1}; 
   case 'xl'
    xL = varargin{i+1};
   case 'xr'
    xR = varargin{i+1};
   case 'nint'
    nint = varargin{i+1}; 
    xbkpt = linspace( xL, xR, nint+1 ); 
   case { 'epsilon', 'eps' } 
    eps = varargin{i+1};
   case 'kord'
    kord = varargin{i+1};
   case 'ncc'
    ncc = varargin{i+1};
   case { 'xbkpt', 'xgrid' }
    xbkpt = varargin{i+1};
   case 'tgrid'
    tGrid = varargin{i+1};
   case 'mf'
    mf = varargin{i+1};
   case 'ode_solver'
    switch lower(varargin{i+1})
     case 'adams'
      METH=1; 
     case 'bdf'
      METH=2; 
     otherwise
      error( 'Unknown argument to ode_solver' ); 
    end
    mf = 10*METH + MITER;
   case 'jacobian'
    switch lower(varargin{i+1})
     case 'exact'
      MITER=1; 
     case { 'approx', 'approximate' }
      MITER=2; 
     otherwise
      error( 'Unknown argument to jacobian' ); 
    end
    mf = 10*METH + MITER;
   case 'dfdu'
    dfdu_func = varargin{i+1};
   case 'dfdux'
    dfdux_func = varargin{i+1};
   case 'dfduxx'
    dfduxx_func = varargin{i+1};
   otherwise
    error( [ 'Unknown input: ', arg ] ); 
  end
  i=i+2;
end

% Input argument error checking:
%
if( xL >= xR ) 
  error( 'xL >= xR' ); 
end
if( (xL ~= xbkpt(1)) | (xR ~= xbkpt(end)) )
  error( 'xL ~= xbkpt(1) OR xR ~= xbkpt(end)' ); 
end
if( eps <= 0 )
  error( 'eps <= 0' ); 
end
if( kord < 3 | kord > 20 )
  error( 'kord < 3 | kord > 20' ); 
end
if( ncc <= 1 | ncc > kord )
  error( 'ncc <= 1 | ncc > kord' ); 
end
if( nint < 1 )
  error( 'nint < 1' ); 
end
if( length(tGrid) < 2 ) 
  error( 'length(tGrid) < 2' ); 
end
% Check if derivative functions were supplied: 
%
exsum = exist('dfdu_func') + exist('dfdux_func') + exist('dfduxx_func'); 
if( ~((exsum==0) | (exsum==3)) )
  error( 'If one of dfdu, dfdux, dfduxx is defined they all must be' ); 
end

% Call the octave *.oct file that calls the FORTRAN routine pdecol:
%
if( exsum==0 ) 
  U = pdecolg(tGrid,dt,xbkpt,eps,kord,ncc,mf,...
	      f_func, dbdu_func, dbdux_func, dzdt_func, uinit_func );
else
  U = pdecolg(tGrid,dt,xbkpt,eps,kord,ncc,mf,...
	      f_func, dbdu_func, dbdux_func, dzdt_func, uinit_func, ...
	      dfdu_func, dfdux_func, dfduxx_func );
end  

% Collect all inputs together into a structure:
%
opts.t0             = t0;
opts.xL             = xL;
opts.xR             = xR; 
opts.nint           = nint; 
opts.dt             = dt;
opts.eps            = eps;
opts.kord           = kord;
opts.ncc            = ncc;
opts.xbkpt          = xbkpt;
opts.tGrid          = tGrid;
opts.mf             = mf;
